from pathlib import Path
from .utils import pass_method, map_values
import numpy as np
from skimage.io import imsave
import os

# from pathlib import Path
from zarr.core import Array


class Patch:
    def __init__(
        self,
        slide_path: str,
        slide_object: Array,
        manager,
        coordinates,
        level: int,
        size: tuple,
        output_suffix: str = "_patch@{}:{}.png",
    ) -> None:
        """
        Init for Patch.
        @param slide_path: Path to slide. Used primarily for generating patch filenames.
        @param slide_object: OpenSlide object. Read patch images from this object.
        @param manager: PatchManager object. Inheriting this allows this patch to be checked for validity.
        @param coordinates: Ndarray of [x, y] coordinates on slide for the top-left corner of the patch.
        @param level: Level of slide you want to call the patch from.
        @param size: tuple of ints for the side lengths of the patch.
        @param output_suffix: The format to be appended onto the slide's name.
        """
        self.manager = manager
        self._slide_path = slide_path
        self.slide_object = slide_object
        self.subfolder = Path(slide_path).stem
        self.coordinates = coordinates
        self.level = level
        self.size = size
        self.output_suffix = output_suffix

    def read_patch(self):
        """
        Read patch from self.slide_object given this patch's coordinates, level, and size.
        @return: PIL object of RGBA patch image.
        """
        return np.asarray(
            self.slide_object.read_region(
                (self.coordinates[1], self.coordinates[0]), self.level, self.size
            )
        )

    def copy(self):
        """
        Return a copy of the current patch.
        @return: new Patch object with same attributes as this one.
        """
        return Patch(
            slide_path=self._slide_path,
            slide_object=self.slide_object,
            manager=self.manager,
            coordinates=self.coordinates,
            level=self.level,
            size=self.size,
        )

    def set_slide(self, slide_path):
        """
        Setter method for changing the slide that this patch belongs to. Useful for pulling corresponding patch from
        multiple slides/label maps.
        @param slide_path: Path of new slide you want to assign this patch to.
        """
        # Change slide path
        self._slide_path = slide_path
        # Re-assign subfolder within output folder
        self.subfolder = Path(slide_path).stem

    def get_patch_path(self, out_dir, create_dir=True):
        """
        Returns string of the path to where this patch will be saved.
        @param out_dir: The output directory
        @return: str
        """
        path = Path(self._slide_path)
        Path(out_dir, self.subfolder).mkdir(parents=True, exist_ok=True)
        return os.path.join(
            out_dir,
            self.subfolder,
            path.name.split(path.suffix)[0]
            + self.output_suffix.format(self.coordinates[0], self.coordinates[1]),
        )

    def save(
        self,
        out_dir,
        save=True,
        check_if_valid=True,
        process_method=None,
        value_map=None,
    ):
        """
        Save patch.
        @param out_dir: Output directory for saving the patch. Supplied by patch_manager.py
        @param save: If False it will not write to disk. Helpful for debugging.
        @param check_if_valid: Run through checks supplied by manager. If rejected, don't save.
        @param process_method: A method that takes an image as an input and returns a string. Summarizes patch info.
            If left as None, it will use utils.pass_method() and return an empty string.
        @param value_map: Map key values in patch to alternate value. dict(key => value) where key, value are ints.
            alters the patch by substituting key for value in the image, leaves values not in dictionary unaltered.
            Helpful for standardization.
        @return: A list [bool, Patch, summary]. bool is if the patch was accepted, Patch is the patch object, and
            summary is the output of process_method.
        """

        patch = self.read_patch()

        if process_method is None:
            process_method = pass_method

        if check_if_valid:
            for check_function in self.manager.valid_patch_checks:
                if not check_function(patch):
                    print("Patch failed check", check_function)
                    return [False, self, ""]

        try:
            if save:
                if isinstance(value_map, dict):
                    patch = self.read_patch()[:, :, 0]
                    patch = map_values(patch, value_map)
                    imsave(fname=self.get_patch_path(out_dir), arr=patch)
                elif value_map is None:
                    patch = self.read_patch()
                    imsave(fname=self.get_patch_path(out_dir), arr=patch)

            return [True, self, process_method(patch)]

        except Exception as e:
            print("Exception while saving patch:", e)
            return [False, self, ""]



# -*- coding: utf-8 -*-
"""
Implementation of UNet
"""
from torch.nn import ModuleList

from GANDLF.models.seg_modules.DownsamplingModule import DownsamplingModule
from GANDLF.models.seg_modules.EncodingModule import EncodingModule
from GANDLF.models.seg_modules.DecodingModule import DecodingModule
from GANDLF.models.seg_modules.UpsamplingModule import UpsamplingModule
from GANDLF.models.seg_modules.InitialConv import InitialConv
from GANDLF.models.seg_modules.out_conv import out_conv
from .modelBase import ModelBase


class unet_multilayer(ModelBase):
    """
    This is the standard U-Net architecture : https://arxiv.org/pdf/1606.06650.pdf. The 'residualConnections' flag controls residual connections, the
    Downsampling, Encoding, Decoding modules are defined in the seg_modules file. These smaller modules are basically defined by 2 parameters, the input
    channels (filters) and the output channels (filters), and some other hyperparameters, which remain constant all the modules. For more details on the
    smaller modules please have a look at the seg_modules file.
    """

    def __init__(
        self,
        parameters: dict,
        residualConnections=False,
    ):
        """
        Parameters
        ----------
        parameters (dict): A dictionary containing the model parameters.
        residualConnections (bool, optional): A flag to control residual connections in the model, by default False.
        """

        self.network_kwargs = {"res": residualConnections}
        super(unet_multilayer, self).__init__(parameters)

        # Set the depth of the model based on the input parameters
        # If the depth is not set, then set it to 4
        parameters["model"]["depth"] = parameters["model"].get("depth", 4)
        self.depth = self.model_depth_check(parameters)

        # Create the initial convolution layer
        self.ins = InitialConv(
            input_channels=self.n_channels,
            output_channels=self.base_filters,
            conv=self.Conv,
            dropout=self.Dropout,
            norm=self.Norm,
            network_kwargs=self.network_kwargs,
        )

        # Create lists of downsampling, encoding, decoding and upsampling modules
        self.ds = ModuleList([])
        self.en = ModuleList([])
        self.us = ModuleList([])
        self.de = ModuleList([])

        # Create the required number of downsampling, encoding, decoding and upsampling modules
        for i_lay in range(0, self.depth):
            self.ds.append(
                DownsamplingModule(
                    input_channels=self.base_filters * 2 ** (i_lay),
                    output_channels=self.base_filters * 2 ** (i_lay + 1),
                    conv=self.Conv,
                    norm=self.Norm,
                )
            )

            self.us.append(
                UpsamplingModule(
                    input_channels=self.base_filters * 2 ** (i_lay + 1),
                    output_channels=self.base_filters * 2 ** (i_lay),
                    conv=self.Conv,
                    interpolation_mode=self.linear_interpolation_mode,
                )
            )

            self.de.append(
                DecodingModule(
                    input_channels=self.base_filters * 2 ** (i_lay + 1),
                    output_channels=self.base_filters * 2 ** (i_lay),
                    conv=self.Conv,
                    norm=self.Norm,
                    network_kwargs=self.network_kwargs,
                )
            )

            self.en.append(
                EncodingModule(
                    input_channels=self.base_filters * 2 ** (i_lay + 1),
                    output_channels=self.base_filters * 2 ** (i_lay + 1),
                    conv=self.Conv,
                    dropout=self.Dropout,
                    norm=self.Norm,
                    network_kwargs=self.network_kwargs,
                )
            )

        # Create the final convolution layer
        self.out = out_conv(
            input_channels=self.base_filters,
            output_channels=self.n_classes,
            conv=self.Conv,
            norm=self.Norm,
            network_kwargs=self.network_kwargs,
            final_convolution_layer=self.final_convolution_layer,
            sigmoid_input_multiplier=self.sigmoid_input_multiplier,
        )

        # Check if converter_type is passed in model, generally referring to ACS
        if "converter_type" in parameters["model"]:
            self.ins = self.converter(self.ins).model
            self.out = self.converter(self.out).model
            for i_lay in range(0, self.depth):
                self.ds[i_lay] = self.converter(self.ds[i_lay]).model
                self.us[i_lay] = self.converter(self.us[i_lay]).model
                self.de[i_lay] = self.converter(self.de[i_lay]).model
                self.en[i_lay] = self.converter(self.en[i_lay]).model

    def forward(self, x):
        """
        Forward pass of the UNet model.

        Args:
            x (Tensor): Should be a 5D Tensor as [batch_size, channels, x_dims, y_dims, z_dims].

        Returns:
            x (Tensor): Returns a 5D Output Tensor as [batch_size, n_classes, x_dims, y_dims, z_dims].

        """
        # Store intermediate feature maps

        y = []
        y.append(self.ins(x))

        # [downsample --> encode] x num layers
        for i in range(0, self.depth):
            temp = self.ds[i](y[i])
            y.append(self.en[i](temp))

        # Get the feature map at the last (deepest) layer
        x = y[-1]

        # [upsample --> encode] x num layers
        for i in range(self.depth - 1, -1, -1):
            # Upsample the feature map to match the size of the corresponding feature map in the encoder
            x = self.us[i](x)

            # Concatenate with the corresponding feature map from the encoder
            x = self.de[i](x, y[i])

        # Get the final output
        x = self.out(x)

        return x


class resunet_multilayer(unet_multilayer):
    """
    This is the standard U-Net architecture with residual connections : https://arxiv.org/pdf/1606.06650.pdf.
    The 'residualConnections' flag controls residual connections, the
    Downsampling, Encoding, Decoding modules are defined in the seg_modules file. These smaller modules are basically defined by 2 parameters, the input
    channels (filters) and the output channels (filters), and some other hyperparameters, which remain constant all the modules. For more details on the
    smaller modules please have a look at the seg_modules file.
    """

    def __init__(self, parameters: dict):
        """
        Parameters
        ----------
        parameters (dict): A dictionary containing the model parameters.

        """
        super(resunet_multilayer, self).__init__(parameters, residualConnections=True)
